// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "File.hpp"

using namespace Dawn;

std::string File::normalizeSlashes(std::string str) {
  size_t i = 0;
  while(i < str.size()) {
    auto c = str[i];
    if(c == '\\' || c == '/') str[i] = FILE_PATH_SEP;
    ++i;
  }
  return str;
}

void File::mkdirp(std::string path) {
  std::string buffer;
  char c;
  size_t i = 0;
  bool inFile;
  bool hasMore;

  inFile = false;
  hasMore = false;
  while(c = path[i]) {
    if((c == '\\' || c == '/') && i > 0) {
      fileMkdir(buffer.c_str(), 0755);
      inFile = false;
      hasMore = false;
      buffer += FILE_PATH_SEP;
      i++;
      continue;
    }

    if(c == '.') inFile = true;
    hasMore = true;
    buffer += c;
    i++;
  }

  if(!inFile && hasMore) {
    fileMkdir(buffer.c_str(), 0755);
  }
}

//

File::File(std::string filename) {
  this->filename = File::normalizeSlashes(filename);
}

bool_t File::open(enum FileMode mode) {
  assertNull(this->file);

  this->mode = mode;
  this->file = fopen(
    this->filename.c_str(),
    mode == FILE_MODE_READ ? "rb" : "wb"
  );

  if(this->file == NULL) return false;

  if(mode == FILE_MODE_READ) {
    fseek(this->file, 0, SEEK_END);
    this->length = ftell(this->file);
    fseek(this->file, 0, SEEK_SET);

    if(this->length <= 0) {
      this->close();
      return false;
    }
  } else {
    this->length = 0;
  }

  return true;
}

bool_t File::isOpen() {
  return this->file != nullptr;
}

bool_t File::exists() {
  if(this->file != nullptr) return true;
  FILE *f = fopen(this->filename.c_str(), "rb");
  if(f == NULL || f == nullptr) return false;
  fclose(f);
  return true;
}

void File::close() {
  assertNotNull(this->file);
  fclose(this->file);
  this->file = nullptr;
}

bool_t File::mkdirp() {
  File::mkdirp(this->filename);
  return true;
}

bool_t File::readString(std::string *out) {
  assertNotNull(out);

  if(!this->isOpen()) {
    if(!this->open(FILE_MODE_READ)) return false;
  }
  assertTrue(this->mode == FILE_MODE_READ);
  out->clear();

  size_t i = 0;
  char buffer[FILE_BUFFER_SIZE + 1];// +1 for null term
  while(i != this->length) {
    size_t amt = mathMin<size_t>(FILE_BUFFER_SIZE, (this->length - i));
    auto amtRead = fread(buffer, sizeof(char), amt, this->file);
    if(amtRead != amt) return false;
    i += amtRead;
    buffer[amtRead] = '\0';
    out->append(buffer);
  }

  return true;
}

bool_t File::writeString(std::string in) {
  if(!this->isOpen()) {
    if(!this->open(FILE_MODE_WRITE)) return false;
  }
  assertTrue(this->mode == FILE_MODE_WRITE);
  return this->writeRaw((char *)in.c_str(), in.size()) && this->length == in.size();
}

bool_t File::writeRaw(char *data, size_t len) {
  if(!this->isOpen()) {
    if(!this->open(FILE_MODE_WRITE)) return false;
  }
  assertTrue(this->mode == FILE_MODE_WRITE);
  this->length = fwrite(data, sizeof(char_t), len, this->file);
  return true;
}

File::~File() {
  if(this->file != nullptr) this->close();
}